import logging
from pathlib import Path

from zimscraperlib.inputs import compute_descriptions
from zimscraperlib.zim import Creator, metadata

from fcc2zim.build import build_command
from fcc2zim.constants import FCC_LANG_MAP, VERSION
from fcc2zim.context import Context
from fcc2zim.fetch import fetch_command
from fcc2zim.prebuild import prebuild_command

logger = Context.logger
context = Context.get()


class Scraper:
    def __init__(self):
        logger.setLevel(level=logging.DEBUG if context.debug else logging.INFO)

        if not (context.do_fetch + context.do_prebuild + context.do_build):
            context.do_fetch = context.do_prebuild = context.do_build = True

        if not context.zimui_dist.exists():
            raise ValueError(
                f"zimui_dist directory {context.zimui_dist} does not exists"
            )

        self.curriculum_raw = context.build_folder / "curriculum-raw"
        self.curriculum_dist = context.build_folder / "curriculum-dist"

        # Make sure the output directory exists
        context.output_folder.mkdir(parents=True, exist_ok=True)
        context.build_folder.mkdir(parents=True, exist_ok=True)

        if context.language not in FCC_LANG_MAP:
            raise ValueError(f"Unsupported language {context.language}")
        self.fcc_lang = FCC_LANG_MAP[context.language]

        context.description, context.long_description = compute_descriptions(
            context.description, context.description, context.long_description
        )

        if context.main_zip_path and not context.main_zip_path.exists():
            raise ValueError(f"main zip file not found in {context.main_zip_path}")
        self.main_zip_path = (
            context.main_zip_path
            if context.main_zip_path
            else context.build_folder / "main.zip"
        )

        if context.i18n_zip_path and not context.i18n_zip_path.exists():
            raise ValueError(f"i18n zip file not found in {context.i18n_zip_path}")
        self.i18n_zip_path = (
            context.i18n_zip_path
            if context.i18n_zip_path
            else context.build_folder / "i18n.zip"
        )

        # if we do not build the ZIM, we can stop here
        if not context.do_build:
            return

        period = context.start_date.strftime("%Y-%m")
        if context.zim_file:
            self.zim_path = Path(context.zim_file.format(period=period))
            # make sure we were given a filename and not a path
            if Path(self.zim_path.name) != self.zim_path:
                raise ValueError(f"zim_name is not a filename: {context.zim_file}")
        else:
            self.zim_path = Path(f"{context.name}_{period}.zim")

        # build full path
        self.zim_path = context.output_folder / self.zim_path

        if self.zim_path.exists():
            if not (context.force or context.overwrite_existing_zim):
                raise ValueError(f"ZIM file {self.zim_path} already exist.")
            logger.info(f"Removing existing ZIM file {self.zim_path}")
            self.zim_path.unlink()
        else:
            logger.info(f"ZIM path: {self.zim_path}")

        logo_path = Path(__file__).parent.joinpath("assets", "fcc_48.png")
        if not logo_path.exists():
            raise ValueError(f"Logo not found at {logo_path}")

        self.creator = Creator(self.zim_path, "index.html").config_metadata(
            std_metadata=metadata.StandardMetadataList(
                Name=metadata.NameMetadata(context.name),
                Language=metadata.LanguageMetadata(context.language),
                Title=metadata.TitleMetadata(context.title),
                Creator=metadata.CreatorMetadata(context.creator),
                Publisher=metadata.PublisherMetadata(context.publisher),
                Date=metadata.DateMetadata(context.start_date),
                Illustration_48x48_at_1=metadata.DefaultIllustrationMetadata(
                    logo_path.read_bytes()
                ),
                Description=metadata.DescriptionMetadata(context.description),
                LongDescription=(
                    metadata.LongDescriptionMetadata(context.long_description)
                    if context.long_description
                    else None
                ),
                Tags=metadata.TagsMetadata(["FCC", "freeCodeCamp"]),
                Scraper=metadata.ScraperMetadata(f"fcc2zim v{VERSION}"),
            )
        )

        # start creator early to detect any problem early as well
        self.creator.start()

    def run(self):
        try:
            self.run_commands()
        except Exception as exc:
            if self.creator:
                self.creator.can_finish = False
            if isinstance(exc, KeyboardInterrupt):
                logger.error("KeyboardInterrupt, exiting.")
                raise SystemExit(3) from exc
            else:
                logger.error(f"Interrupting process due to error: {exc}")
                logger.exception(exc)
                raise SystemExit(2) from exc
        else:
            if self.creator:
                self.creator.finish()
                logger.info(f"Finished creating ZIM at {self.zim_path}")

    def run_commands(self):
        if context.do_fetch:
            fetch_command(
                force=context.force,
                curriculum_raw=self.curriculum_raw,
                zip_main_path=self.main_zip_path,
                zip_i18n_path=self.i18n_zip_path,
                fcc_lang=self.fcc_lang,
            )
        if context.do_prebuild:
            prebuild_command(
                fcc_lang=self.fcc_lang,
                course_list=context.course,
                curriculum_raw=self.curriculum_raw,
                curriculum_dist=self.curriculum_dist,
            )
        if context.do_build:
            build_command(
                creator=self.creator,
                zimui_dist=context.zimui_dist,
                curriculum_dist=self.curriculum_dist,
            )



import shutil
import zipfile
from pathlib import Path

import requests

from fcc2zim.context import Context

logger = Context.logger


def fetch_command(
    zip_main_path: Path,
    zip_i18n_path: Path,
    curriculum_raw: Path,
    fcc_lang: str,
    *,
    force: bool,
):
    logger.info("Scraper: fetch phase starting")
    main_url = (
        "https://github.com/freeCodeCamp/freeCodeCamp/archive/refs/heads/main.zip"
    )
    i18n_url = (
        "https://github.com/freeCodeCamp/i18n-curriculum/archive/refs/heads/main.zip"
    )

    # Don't redownload the file if we already have it (it's a large file)
    for zip_path in [zip_main_path, zip_i18n_path]:
        if force or not zip_path.exists():
            logger.debug(f"Download zip file to {zip_path}")
            resp = requests.get(
                main_url if zip_path == zip_main_path else i18n_url,
                allow_redirects=True,
                timeout=5,
            )
            zip_path.write_bytes(resp.content)
        else:
            logger.debug(f"Using existing zip file {zip_path}")

    curriculum_raw.mkdir(parents=True, exist_ok=True)
    shutil.rmtree(curriculum_raw)

    logger.debug("Extracting files")
    for zip_path in [zip_main_path, zip_i18n_path]:
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            count = 0
            for member in zip_ref.namelist():
                if not member.endswith("/") and (
                    member.startswith("freeCodeCamp-main/curriculum/challenges/_meta")
                    or member.startswith(
                        f"freeCodeCamp-main/curriculum/challenges/{fcc_lang}"
                    )
                    or member.startswith(
                        f"freeCodeCamp-main/curriculum/dictionaries/{fcc_lang}"
                    )
                    or member.startswith(
                        f"freeCodeCamp-main/client/i18n/locales/{fcc_lang}"
                    )
                    or member.startswith(
                        f"i18n-curriculum-main/curriculum/challenges/{fcc_lang}"
                    )
                    or member.startswith(
                        f"i18n-curriculum-main/curriculum/dictionaries/{fcc_lang}"
                    )
                ):
                    # compute path to merge freeCodeCamp and i18n-curriculum content
                    target_path = (
                        curriculum_raw
                        / "extracted"
                        / member[18 if member.startswith("freeCodeCamp-main") else 21 :]
                    )
                    target_path.parent.mkdir(parents=True, exist_ok=True)
                    target_path.write_bytes(zip_ref.read(member))
                    count += 1
            logger.info(f"Extracted {count} files from {zip_path}")
    logger.info(f"Fetched curriculum into {curriculum_raw}")
    logger.info("Scraper: fetch phase finished")



import dataclasses
import datetime
import logging
import os
from pathlib import Path
from typing import Any

from zimscraperlib.logging import DEFAULT_FORMAT_WITH_THREADS, getLogger

from fcc2zim.constants import (
    NAME,
)


@dataclasses.dataclass(kw_only=True)
class Context:
    """Class holding every contextual / configuration bits which can be moved

    Used to easily pass information around in the scraper. One singleton instance is
    always available.
    """

    # singleton instance
    _instance: "Context | None" = None

    # list of courses to include
    course: list[str]

    # ZIM language, also used to fetch proper FCC content
    language: str

    # ZIM name (also used for filename if zim_file is not set) and filename
    name: str
    zim_file: str | None = None

    # Other ZIM metadata
    title: str
    description: str
    long_description: str | None = None
    creator: str = "freeCodeCamp"
    publisher: str = "openZIM"

    # force flag
    force: bool = False

    # debug flag
    debug: bool = False

    # do flags (to control which steps will be done)
    do_fetch: bool = os.getenv("DO_FETCH", "False").lower() == "true"
    do_prebuild: bool = os.getenv("DO_PREBUILD", "False").lower() == "true"
    do_build: bool = os.getenv("DO_BUILD", "False").lower() == "true"

    # folder where the ZIM will be written
    output_folder: Path = Path(os.getenv("FCC_OUTPUT", "../output"))

    # folder where temporary files will be stored
    build_folder: Path = Path(os.getenv("FCC_BUILD", "../build"))

    # folder where Vue.JS UI has been built
    zimui_dist: Path = Path(os.getenv("FCC_ZIMUI_DIST", "../zimui/dist"))

    # ZIP of FCC content
    main_zip_path: Path | None = None
    i18n_zip_path: Path | None = None

    # Do not fail if ZIM already exists, overwrite it
    overwrite_existing_zim: bool = False

    # Date when the crawl started
    start_date: datetime.date

    # logger to use everywhere (do not mind about mutability, we want to reuse same
    # logger everywhere)
    logger: logging.Logger = getLogger(  # noqa: RUF009
        NAME, level=logging.DEBUG, log_format=DEFAULT_FORMAT_WITH_THREADS
    )

    @classmethod
    def setup(cls, **kwargs: Any):
        new_instance = cls(**kwargs)
        if cls._instance:
            # replace values 'in-place' so that we do not change the Context object
            # which might be already imported in some modules
            for field in dataclasses.fields(new_instance):
                cls._instance.__setattr__(
                    field.name, new_instance.__getattribute__(field.name)
                )
        else:
            cls._instance = new_instance

    @classmethod
    def get(cls) -> "Context":
        if not cls._instance:
            raise OSError("Uninitialized context")  # pragma: no cover
        return cls._instance



import pathlib
import re

import yaml

FRONTMATTER_BOUNDARY = re.compile(r"^-{3,}\s*$", re.MULTILINE)


# Each markdown challenge contains 'frontmatter' which is a header
# consisting of yaml formatted data enclosed inside of '---\n' lines
# We load into YAML only the frontmatter
def read_yaml_frontmatter(filename: pathlib.Path) -> dict[str, str]:
    _, fm, _ = FRONTMATTER_BOUNDARY.split(filename.read_text("UTF8"), 2)
    return yaml.load(  # pyright: ignore[reportReturnType, reportUnknownVariableType, reportUnknownMemberType]
        fm, Loader=yaml.SafeLoader
    )


class Challenge:
    def __init__(self, fpath: str | pathlib.Path) -> None:
        self.path = pathlib.Path(fpath)
        self.course_slug = self.path.parent.stem
        self.course_superblock = "-".join(self.path.parent.parent.stem.split("-")[1:])
        self.language = self.path.parent.parent.parent.stem
        self._frontmatter = None

    def identifier(self):
        return str(self.frontmatter()["id"])

    def title(self):
        return str(self.frontmatter()["title"])

    def frontmatter(self):
        if not self._frontmatter:
            self._frontmatter = read_yaml_frontmatter(self.path)
        return self._frontmatter



import argparse
import datetime
from pathlib import Path

from zimscraperlib.constants import (
    MAXIMUM_DESCRIPTION_METADATA_LENGTH as MAX_DESC_LENGTH,
)
from zimscraperlib.constants import (
    MAXIMUM_LONG_DESCRIPTION_METADATA_LENGTH as MAX_LONG_DESC_LENGTH,
)
from zimscraperlib.constants import RECOMMENDED_MAX_TITLE_LENGTH

from fcc2zim.constants import FCC_LANG_MAP, NAME, VERSION
from fcc2zim.context import Context


def prepare_context(raw_args: list[str]) -> None:
    """Initialize scraper context from command line arguments"""

    parser = argparse.ArgumentParser(
        prog=NAME,
        description="Scraper to create ZIM files from freeCodeCamp courses",
    )

    parser.add_argument(
        "--course",
        help="Course or course list (separated by commas)",
        type=lambda x: [course.strip() for course in x.split(",")],
        required=True,
    )

    parser.add_argument(
        "--language",
        type=str,
        help="Curriculum language",
        required=True,
        choices=FCC_LANG_MAP.keys(),
    )

    parser.add_argument(
        "--name",
        type=str,
        help="ZIM name. Used as identifier and filename (date will be appended)",
        required=True,
    )

    parser.add_argument(
        "--title",
        type=str,
        help=f"Title of ZIM file (less than {RECOMMENDED_MAX_TITLE_LENGTH} graphemes)",
        required=True,
    )

    parser.add_argument(
        "--description",
        type=str,
        help=f"Description of ZIM file (less than {MAX_DESC_LENGTH} graphemes)",
        required=True,
    )

    parser.add_argument(
        "--long-description",
        type=str,
        help=f"Long description of ZIM file (less than {MAX_LONG_DESC_LENGTH} "
        "graphemes)",
    )

    parser.add_argument(
        "--creator",
        type=str,
        help=f"Name of freeCodeCamp courses creator. Default: {Context.creator!s}",
    )

    parser.add_argument(
        "--publisher",
        type=str,
        help=f"Publisher of the zim file. Default: {Context.publisher!s}",
    )

    parser.add_argument(
        "--force",
        help="Force a full reprocessing, not benefiting from any cached file",
        action="store_true",
    )

    parser.add_argument(
        "--debug",
        help="Enable verbose output",
        action="store_true",
    )

    parser.add_argument(
        "--output",
        type=Path,
        help="Output directory where zim file will be built",
        dest="output_folder",
    )

    parser.add_argument(
        "--build",
        type=Path,
        help="The build directory to hold temporary files during scraper operation",
        dest="build_folder",
    )

    parser.add_argument(
        "--zimui-dist",
        type=Path,
        help=(
            "Dev option to customize directory containing Vite build output from the "
            "ZIM UI Vue.JS application"
        ),
    )

    parser.add_argument(
        "--zim-file",
        type=str,
        help="ZIM file name (based on --name if not provided), could contain {period}"
        " placeholder which will be replaced by <year>_<month>",
    )

    parser.add_argument(
        "--zip-path",
        help="Path to main zip file containing FCC courses",
        type=str,
        dest="main_zip_path",
    )

    parser.add_argument(
        "--i18n-zip-path",
        help="Path to i18n zip file containing translations of FCC courses",
        type=str,
    )

    parser.add_argument(
        "--version",
        help="Display scraper version and exit",
        action="version",
        version=f"fcc2zim {VERSION}",
    )

    parser.add_argument(
        "--overwrite",
        help="Do not fail if ZIM already exists, overwrite it",
        action="store_true",
        dest="overwrite_existing_zim",
    )

    args = parser.parse_args(raw_args)

    # Ignore unset values so they do not override the default specified in Context
    args_dict = {key: value for key, value in args._get_kwargs() if value}

    args_dict["start_date"] = datetime.date.today()

    Context.setup(**args_dict)



import pathlib

from fcc2zim.__about__ import __version__

NAME = "fcc2zim"
VERSION = __version__
ROOT_DIR = pathlib.Path(__file__).parent

# key is the language passed at CLI (and used as ZIM Language metadata)
# value is the name of folders used in FCC source code
FCC_LANG_MAP = {
    # removed until we settle on these two codes
    # "cmn": "chinese",
    # "lzh": "chinese-traditional",
    "eng": "english",
    "spa": "espanol",
    "deu": "german",
    "ita": "italian",
    "jpn": "japanese",
    "por": "portuguese",
    "ukr": "ukrainian",
    "swa": "swahili",
}



import tempfile
from pathlib import Path

from fcc2zim.challenge import read_yaml_frontmatter


def test_good_front_matter():
    with tempfile.TemporaryDirectory() as tmp_dir:
        md_file = Path(tmp_dir) / "file.md"
        # dd:ee\ndd: is bad YAML which is intentionaly here, because outside
        # the front matter and should hence be properly ignored
        md_file.write_text("----\ntitle: foo\nsubject: bar\n---\ndd:ee\ndd:")
        front = read_yaml_frontmatter(md_file)
        assert front["title"] == "foo"
        assert front["subject"] == "bar"



import json
from dataclasses import dataclass
from pathlib import Path

from zimscraperlib.zim import Creator

from fcc2zim.context import Context

logger = Context.logger


@dataclass
class CurriculumRedirect:
    path: str
    title: str


def build_curriculum_redirects(curriculum_dist: Path):
    """
    Build the list of redirects from challenge URL to Vite hash URL

    The Vite app uses its own router to navigate. We have a single HTML file, but we
    need an URL for each challenge for the zim search to work.
    This builds the list of redirect needed fron the challenge URL to Vite hash URL.
    """
    index_json_path = curriculum_dist.joinpath("curriculum", "index.json")
    with open(index_json_path) as course_index_str:
        superblock_dict = json.load(course_index_str)

    redirects: list[CurriculumRedirect] = []
    for superblock in superblock_dict:
        course_list = superblock_dict[superblock]
        for course in course_list:
            meta_json_path = Path(
                curriculum_dist,
                "curriculum",
                superblock,
                course,
                "_meta.json",
            )
            challenges = json.loads(meta_json_path.read_text())["challenges"]
            for challenge in challenges:
                title = challenge["title"]
                redirects.append(
                    CurriculumRedirect(
                        path=f'{superblock}/{course}/{challenge["slug"]}', title=title
                    )
                )

    return redirects


def build_command(
    zimui_dist: Path,
    creator: Creator,
    curriculum_dist: Path,
):
    logger.info("Scraper: build phase starting")

    mathjax = (Path(__file__) / "../mathjax").resolve()
    count_mathjax_files = len(list(mathjax.rglob("*")))
    logger.info(f"Adding {count_mathjax_files} MathJax files in {mathjax}")
    for file in mathjax.rglob("*"):
        if not file.is_file():
            continue
        path = str(Path(file).relative_to(mathjax.parent))
        logger.debug(f"Adding {path} to ZIM")
        creator.add_item_for(
            path=path,
            fpath=file,
            is_front=False,
        )

    # Add zimui files
    for file in zimui_dist.rglob("*"):
        if file.is_dir():
            continue
        path = str(Path(file).relative_to(zimui_dist))
        logger.debug(f"Adding {path} to ZIM")
        creator.add_item_for(path, fpath=file)

    # Add prebuild generated curriculum file
    for file in curriculum_dist.rglob("*"):
        if file.is_dir():
            continue
        path = str(Path("content").joinpath(Path(file).relative_to(curriculum_dist)))
        logger.debug(f"Adding {path} to ZIM")
        creator.add_item_for(path, fpath=file)

    for redirection in build_curriculum_redirects(curriculum_dist=curriculum_dist):
        redirect_path = f"index/{redirection.path}.html"
        redirect_url = (
            redirection.path.count("/") + 1
        ) * "../" + f"index.html#{redirection.path}"
        content = (
            f"<html><head><title>{redirection.title}</title>"
            f'<meta http-equiv="refresh" content="0;URL=\'{redirect_url}\'" />'
            f"</head><body></body></html>"
        )
        logger.debug(
            f"Redirecting {redirect_path} to {redirect_url} for slug {redirection.path}"
            f"and title {redirection.title}",
        )
        creator.add_item_for(
            redirect_path,
            content=bytes(content, "utf-8"),
            title=redirection.title,
            mimetype="text/html",
            is_front=True,
        )
        # Example index.html#/regular-expressions/extract-matches

    logger.info("Scraper: build phase finished")



import sys

from fcc2zim.context import Context
from fcc2zim.entrypoint import prepare_context

logger = Context.logger


def main():
    try:
        prepare_context(sys.argv[1:])

        # import this only once the Context has been initialized, so that it gets an
        # initialized context
        from fcc2zim.scraper import Scraper

        Scraper().run()

    except SystemExit:
        logger.error("Scraper failed, exiting")
        raise
    except Exception as exc:
        logger.exception(exc)
        logger.error(f"Scraper failed with the following error: {exc}")
        raise SystemExit(1) from exc

    logger.info("Scraper completed")


if __name__ == "__main__":
    main()
